import torch
import numpy as np

class NoiseMixer():
    def __init__(self, augment_prob=0.2, utterance_mix_ratio=0.25, shift_range=[0.4,0.7],
                 magnitude_range=[0.05,0.7],utterance_magnitude_max_scale=0.2):
        self.augment_prob = augment_prob
        self.utterance_mix_ratio = utterance_mix_ratio
        self.shift_range = shift_range
        self.magnitude_range = magnitude_range
        self.utterance_magnitude_max_scale=utterance_magnitude_max_scale

    def __call__(self, wav, noise):
        # wav: (B,L,)
        # noise: (B,L,)

        is_augmented = (torch.rand_like(wav[:,0])<= self.augment_prob)*1.0 # (B,)
        is_utterance_mix = (torch.rand_like(wav[:,0])<= self.utterance_mix_ratio)*1.0 # (B,)
        idxs = np.arange(len(wav))
        np.random.shuffle(idxs)
        shuffled_wav = wav[idxs]
        
        shift_ranges = torch.rand_like(wav[:,0])*(self.shift_range[1]-self.shift_range[0])+self.shift_range[0]
        left_shift_mask = (torch.linspace(0,1,noise.shape[1], device=noise.device)[None,:] > shift_ranges[:,None])*1.0
        right_shift_mask = (torch.linspace(1,0,noise.shape[1], device=noise.device)[None,:] > shift_ranges[:,None])*1.0
        is_left = (torch.rand_like(wav[:,0])>=0.5)*1.0

        is_utterance_mix = is_utterance_mix[:,None]
        is_left = is_left[:,None]
        
        noise = (1-is_utterance_mix)*noise + is_utterance_mix*(is_left*left_shift_mask*shuffled_wav+(1-is_left)*right_shift_mask*shuffled_wav)        
        magnitude = torch.rand_like(wav[:,0])*(self.magnitude_range[1]-self.magnitude_range[0])+self.magnitude_range[0]
        utt_magnitude = torch.rand_like(wav[:,0])*(self.utterance_magnitude_max_scale-self.magnitude_range[0])+self.magnitude_range[0]
        magnitude = utt_magnitude*is_utterance_mix.squeeze(-1) + (1-is_utterance_mix.squeeze(-1))*magnitude
        magnitude = is_augmented[:,None]*magnitude[:,None]
        wav = wav+magnitude*noise
        return wav
        


import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from lightning import LightningModule
from torch.optim.lr_scheduler import LambdaLR
from torch.nn import init
from ..utils.lr_schedule import COSLRLAMBDA
from .sylber import Sylber
from .segment_synthesis import SegmentSynthesisTrainer

class SylberTrainer(LightningModule):

    def __init__(self, loss_coefs,lr=0.0001, warmup_steps=5000, total_steps=500000,
                 min_factor=1.0,hold_steps=0, accumulate_grad_batches=1,
                 **model_configs):
        super().__init__()
        self.loss_coefs = loss_coefs
        
        self.net = Sylber(**model_configs).to(torch.float)
        self.lr = lr
        self.total_steps =total_steps
        self.warmup_steps = warmup_steps
        self.min_factor = min_factor
        self.hold_steps = hold_steps
        self.accumulate_grad_batches = accumulate_grad_batches
        
    def forward(self, **kwargs):
        return self.net(**kwargs)
        
    def training_step(self, batch, batch_idx):
        
        if (self.global_step%self.accumulate_grad_batches) ==0:
            self.net.ema_step()
        outputs = self.net(**batch)
        
        loss_val = 0
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'train_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'train_loss', loss_val)
        if self.net.thresholder is not None:
            self.log('normthreshold', self.net.thresholder.get_threshold().item())
        return loss_val

        
    def validation_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        loss_val = 0
        
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'val_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'val_loss', loss_val,sync_dist=True)
            
        return loss_val

    def configure_optimizers(self):
        
        opt_fun = torch.optim.AdamW
        opt = opt_fun(self.net.parameters(),lr=self.lr,eps=1e-4, betas=(0.9, 0.95), weight_decay = 0.1)
        lr_lambda = COSLRLAMBDA(self.warmup_steps, self.total_steps, self.min_factor,  self.hold_steps)
        sch = LambdaLR(opt, lr_lambda)
        return [opt],[{"scheduler": sch, "interval": "step"}]

import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from lightning import LightningModule
from torch.optim.lr_scheduler import LambdaLR
from torch.nn import init
from ..utils.lr_schedule import COSLRLAMBDA
from .sylber import Sylber
from .segment_synthesis import SegmentSynthesis


class SylberTrainer(LightningModule):

    def __init__(self, loss_coefs,lr=0.0001, warmup_steps=5000, total_steps=500000,
                 min_factor=1.0,hold_steps=0, accumulate_grad_batches=1,
                 **model_configs):
        super().__init__()
        self.loss_coefs = loss_coefs
        
        self.net = Sylber(**model_configs).to(torch.float)
        self.lr = lr
        self.total_steps =total_steps
        self.warmup_steps = warmup_steps
        self.min_factor = min_factor
        self.hold_steps = hold_steps
        self.accumulate_grad_batches = accumulate_grad_batches
        
    def forward(self, **kwargs):
        return self.net(**kwargs)
        
    def training_step(self, batch, batch_idx):
        
        if (self.global_step%self.accumulate_grad_batches) ==0:
            self.net.ema_step()
        outputs = self.net(**batch)
        
        loss_val = 0
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'train_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'train_loss', loss_val)
        if self.net.thresholder is not None:
            self.log('normthreshold', self.net.thresholder.get_threshold().item())
        return loss_val

        
    def validation_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        loss_val = 0
        
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'val_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'val_loss', loss_val,sync_dist=True)
            
        return loss_val

    def configure_optimizers(self):
        
        opt_fun = torch.optim.AdamW
        opt = opt_fun(self.net.parameters(),lr=self.lr,eps=1e-4, betas=(0.9, 0.95), weight_decay = 0.1)
        lr_lambda = COSLRLAMBDA(self.warmup_steps, self.total_steps, self.min_factor,  self.hold_steps)
        sch = LambdaLR(opt, lr_lambda)
        return [opt],[{"scheduler": sch, "interval": "step"}]


class SegmentSynthesisTrainer(LightningModule):

    def __init__(self, loss_coefs,lr=0.0001, warmup_steps=5000, total_steps=500000, min_factor=0.05, 
                 **model_configs):
        super().__init__()
        self.loss_coefs = loss_coefs
        
        self.net = SegmentSynthesis(**model_configs).to(torch.float)
        self.lr = lr
        self.total_steps =total_steps
        self.warmup_steps = warmup_steps
        self.min_factor = min_factor
        
    def forward(self, **kwargs):
        return self.net(**kwargs)
    
    def training_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        
        loss_val = 0
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'train_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'train_loss', loss_val)
        
        return loss_val

        
    def validation_step(self, batch, batch_idx):
        outputs = self.net(**batch)
        loss_val = 0
        
        for coef_name, coef_val in self.loss_coefs.items():
            if coef_name in outputs.keys():
                loss_val += coef_val * outputs[coef_name]
                self.log(f'val_{coef_name}', outputs[coef_name],sync_dist=True)
        self.log(f'val_loss', loss_val,sync_dist=True)
            
        return loss_val

    def configure_optimizers(self):
        
        opt_fun = torch.optim.AdamW
        opt = opt_fun(self.net.parameters(),lr=self.lr,eps=1e-4, betas=(0.9, 0.95), weight_decay = 0.1)
        lr_lambda = COSLRLAMBDA(self.warmup_steps, self.total_steps, self.min_factor)
        sch = LambdaLR(opt, lr_lambda)
        return [opt],[{"scheduler": sch, "interval": "step"}]



import math

class LRLAMBDA(object):
    def __init__(self, warmup_steps, total_steps, min_factor=0.05, hold_steps=0 ):
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self.min_factor = min_factor
        self.hold_steps = hold_steps
        
    
    def __call__(self, step):
        if step < self.warmup_steps:
            return float(step) / float(max(1.0, self.warmup_steps)) * 1
        else:
            net_step = max(0,step-self.warmup_steps-self.hold_steps)

            return max(
                self.min_factor, 1.0 - float(net_step) / float(max(1, net_step))
            ) * 1
            

class COSLRLAMBDA(object):
    def __init__(self, warmup_steps, total_steps, min_factor=0.05, hold_steps=0 ):
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self.min_factor = min_factor
        self.hold_steps = hold_steps
        
    
    def __call__(self, step):
        if step < self.warmup_steps:
            return float(step) / float(max(1.0, self.warmup_steps)) * 1
        elif step> (self.total_steps + self.hold_steps+self.warmup_steps):
            return self.min_factor
        else:
            net_step = max(0,step-self.warmup_steps-self.hold_steps)
            return self.min_factor+(1-self.min_factor)*(1+math.cos( float(net_step)/float(max(1, self.total_steps)) *math.pi))/2



import numpy as np
import torch
import torch.nn as nn


class Thresholder(nn.Module):
    def __init__(self, signal_mean=None, signal_var=None, noise_mean=None, noise_var=None, decay=0.9999, eta=1,
                threshold=None):
        super().__init__()
        if threshold is None:
            self.signal_mean = nn.parameter.Parameter(torch.ones(1)*signal_mean,requires_grad=False)
            self.signal_var = nn.parameter.Parameter(torch.ones(1)*signal_var,requires_grad=False)
            self.noise_mean = nn.parameter.Parameter(torch.ones(1)*noise_mean,requires_grad=False)
            self.noise_var = nn.parameter.Parameter(torch.ones(1)*noise_var,requires_grad=False)
            self.decay = decay
            self.eta = eta
            self.threshold =  None
        else:
            self.signal_mean = nn.parameter.Parameter(torch.ones(1),requires_grad=False)
            self.signal_var = nn.parameter.Parameter(torch.ones(1),requires_grad=False)
            self.noise_mean = nn.parameter.Parameter(torch.ones(1),requires_grad=False)
            self.noise_var = nn.parameter.Parameter(torch.ones(1),requires_grad=False)
            self.decay = decay
            self.eta = eta
            self.threshold =  nn.parameter.Parameter(torch.ones(1)*threshold,requires_grad=False)
        
    def get_threshold(self):
        if self.threshold is not None:
            return self.threshold
        else:
            with torch.no_grad():
                
                mu_S = self.signal_mean  # Mean of noise
                sigma_S = (self.signal_var+1e-8)**.5  # Standard deviation of noise
                mu_N = self.noise_mean  # Mean of signal
                sigma_N = (self.noise_var+1e-8)**.5  # Standard deviation of signal
                eta = self.eta
                A = sigma_S**2 - sigma_N**2
                B = -2*sigma_S**2*mu_N + 2*sigma_N**2*mu_S
                C = sigma_S**2*mu_N**2 - sigma_N**2*mu_S**2 - 2*sigma_N**2*sigma_S**2*(np.log(eta) + torch.log(sigma_S/sigma_N))
    
                if A != 0:
                    discriminant = B**2 - 4*A*C
                    if discriminant > 0:
                        threshold = (-B + ((mu_S>mu_N)*1.0)*torch.sqrt(discriminant)) / (2*A)
                    elif discriminant == 0:
                        threshold = -B / (2*A)
                else:
                    # When A is zero, solve the linear equation Bx + C = 0
                    if B != 0:
                        threshold = -C / B
            return threshold

    
    def update_stats(self, signal, noise):
        if self.threshold is not None:
            return
        with torch.no_grad():
            if signal is not None:
                self.signal_mean.data = self.decay*self.signal_mean.data+(1-self.decay)*signal.mean()
                self.signal_var.data = self.decay*self.signal_var.data+(1-self.decay)*((signal-self.signal_mean.data)**2).mean()
            if noise is not None:
                self.noise_mean.data = self.decay*self.noise_mean.data+(1-self.decay)*noise.mean()
                self.noise_var.data = self.decay*self.noise_var.data+(1-self.decay)*((noise-self.noise_mean.data)**2).mean()



def cossim(x,y):
    return (x*y).sum(-1)/(((x**2).sum(-1)+1e-8)**.5)/(((y**2).sum(-1)+1e-8)**.5)


def get_segment(states, normthreshold, mergethreshold, norms=None):
    # states: (L, d)
    if norms is None:
        norms = ((states**2).sum(-1)+1e-8)**.5
    mask = norms>=normthreshold

    curr = 0
    seg_cnt = 0
    segments = []
    s=-1
    midboundaries = []
    for i in range(len(states)):
        if ~mask[i]:
            seg_cnt = 0
            if s>-1:
                segments.append([s,i])
            s = -1
            curr = 0
            seg_cnt = 0
        else:
            if seg_cnt ==0:
                curr = states[i]
                seg_cnt+=1
                s = i
            else:
                sim = cossim(curr, states[i])
                if sim >= mergethreshold:
                    curr = (curr*seg_cnt+states[i])/(seg_cnt+1)
                    seg_cnt += 1
                else:
                    curr = states[i]
                    seg_cnt += 1
                    segments.append([s,i])
                    midboundaries.append([i,len(segments)-1])
                    s = i
    if s>-1:
        segments.append([s,len(states)])
        
    merged = []
    for bd, segi in midboundaries:
        if segi >= len(segments)-1:
            continue
        if cossim(states[segments[segi][0]:segments[segi][1]].mean(0),
                  states[segments[segi+1][0]:segments[segi+1][1]].mean(0))>=mergethreshold:
            segments[segi+1]=[segments[segi][0],segments[segi+1][1]]
            merged.append(segi)
            continue
        s = max(segments[segi][0],bd-max(1,(segments[segi][1]-segments[segi][0])//2))
        bd = min(segments[segi+1][1],bd+max(1,(segments[segi+1][1]-segments[segi+1][0])//2))
        prev_center = states[segments[segi][0]:segments[segi][1]].mean(0)
        next_center = states[segments[segi+1][0]:segments[segi+1][1]].mean(0)
        sim_prev = cossim(states[s:bd],prev_center[None,:])
        sim_next = cossim(states[s:bd],next_center[None,:])
        sim_sweep = [(sim_prev[:i].sum()+sim_next[i:].sum()) for i in range(0,bd-s)]
        opt_b = np.arange(s,bd)[np.argmax(sim_sweep)]
        segments[segi] = [segments[segi][0],opt_b]
        segments[segi+1] = [opt_b,segments[segi+1][1]]  
    
    segments = [segment for segi, segment in enumerate(segments) if segi not in merged]                
    return np.array(segments)



import numpy as np
import torch
import yaml
from pathlib import Path
from model.sylbe2 import Sylber

def load_model(ckpt_path, **kwargs):
    try:
        ckpt = torch.load(ckpt_path)
        cfg = ckpt['config']
        state_dict = ckpt['state_dict']
    except:
        cfg, ckpt_path = load_cfg_and_ckpt_path(version_dir=ckpt_path,**kwargs)
        ckpt = torch.load(ckpt_path)
        cfg = cfg['trainer']['model_configs']
        state_dict = OrderedDict()
        for module_name, state in state_dict.items():
            if f'net.' in module_name:
                new_name = module_name.split(f'net.')[-1]
                state_dict[new_name] = state
        
    model = Sylber(**cfg)
    model.load_state_dict(state_dict, strict=False)
    model = model.eval()

    return model, cfg


def load_cfg_and_ckpt_path(version_dir, mode='latest'):
    version_dir = Path(version_dir)
    try:
        cfg = yaml.load(open(version_dir / '.hydra' / 'config.yaml'), Loader=yaml.FullLoader)
    except:
        cfg = yaml.load(open(version_dir.parent / '.hydra' / 'config.yaml'), Loader=yaml.FullLoader)
    version_name = [f for f in (version_dir / 'lightning_logs').glob('version_*')][0].stem
    if mode == 'best':
        checkpoint_path = [f for f in Path(version_dir / 'lightning_logs' / version_name / 'checkpoints').glob('*.ckpt')
                           if 'best' in f.name]
        checkpoint_path = checkpoint_path[-1]
    else:
        checkpoint_path = [f for f in
                           Path(version_dir / 'lightning_logs' / version_name / 'checkpoints').glob('*.ckpt')]

        def get_epoch(fileName):
            try:
                epoch = [n for n in fileName.split('-') if 'epoch' in n][0]
            except:
                return -1
            return int(epoch.split('=')[-1])

        checkpoint_path.sort(key=lambda f: get_epoch(f.name))
        checkpoint_path = checkpoint_path[-1]
        
    print(f"{str(checkpoint_path)} is located.")
    return cfg, str(checkpoint_path)